/*
** Module   : ToggleBox.java
** Abstract : toggle box implementation
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20140610 Initial implementation.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of a "toggle box" with a text label.
 */
public abstract class ToggleBox
extends AbstractWidget
implements KeyListener,
           FocusListener, 
           LabeledWidget/*,
           DataContainer*/
{
   /** Button label. */
   private String textLabel;

   /** Container for listeners. */
   private ArrayList<ActionListener> listeners = new ArrayList<ActionListener>();

   /** Toggle Box attribute container. */
   private ToggleBoxConfig config = null;

   /** Index of the character to underscore. */
   protected int underscore = -1;
   
   /** Label instance associated with widget. */
   private Label label = null;
   
   /** Widget in focus flag. */
   protected boolean widgetFocused = false;
   
   /** Previous cursor visibility flag to restore on focus lost. */
   private boolean curvis = false;
   
   /**
    * This constructor creates a toggle-box with a given text label.
    * 
    * @param   owner
    *          The parent container, frame or dialog which will include this toggle-box.
    * @param   text
    *          The string to be used as the toggle-box text.
    */
   public ToggleBox(Container owner, String text)
   {
      setTextLabel(text);
      
      config = new ToggleBoxConfig(ComponentConfig.TOGGLE_BOX, this);
      
      config.setName(text);
      config.setWidthCharsDirect(width());
      config.setHelp(null);
      addFocusListener(this);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   cfg
    *          Configuration data.
    */
   public ToggleBox(ComponentConfig cfg)
   {
      setTextLabel("");
      config = new ToggleBoxConfig(this, cfg);
      config.applyConfig(cfg);

      if (cfg.getLabel() != null)
         setText(cfg.getLabel());
      addFocusListener(this);
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Get label text.
    * 
    * @return  label text.
    */
   public String textLabel()
   {
      return textLabel;
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @Override
   @SuppressWarnings("unchecked")
   public ToggleBoxConfig config()
   {
      return config;
   }

   /**
    * Draw the button. Called by this ToggleBox's parent container.
    */
   @Override
   public abstract void draw();

   /**
    * Return current value of COLON option.
    * 
    * @return  Current value of the option column.
    */
   @Override
   public int getColon()
   {
      return config.getColon();
   }

   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   @Override
   public String getColumnLabelText()
   {
      return null;
   }

   /**
    * Return reference to the instance of the <code>Label</code> attached to
    * the widget.
    * 
    * @return  Reference to <code>Label</code> instance.
    */
   @Override
   public Label getLabelInstance()
   {
      return label;
   }

   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   @Override
   public String getLabelText()
   {
      return null;
   }
   
   /**
    * Return TO column.
    * 
    * @return  Column assigned via TO option.
    */
   @Override
   public int getTo()
   {
      return config.getTo();
   }
   
   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
/*   @Override
   public BaseDataType getValue()
   {
      return new logical(config.isChecked());
   }*/

   /**
    * Calculate the height of this widget.
    *
    * @return    The height in rows.
    */
   @Override
   public int height()
   {
      return config.getHeightChars() > 0 ? config.getHeightChars() : 1;
   }
   
   /**
    * Check if widget has fixed column position.
    * 
    * @return  <code>true</code> if widget column is fixed.
    */
   @Override
   public boolean isColumnFixed()
   {
      return config.getAssignedColumn() != BaseConfig.INV_COORD;
   }

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isForceLabel()
   {
      return config.isForceLabel();
   }

   /**
    * Get current state of NO-LABELS flag.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isNoLabels()
   {
      return config.isNoLabels();
   }

   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      return false;
   }

   /**
    * Calculate the width of this widget.
    *
    * @return    The width in characters.
    */
   @Override
   public int width()
   {
      return config.getWidthChars() > 0 ? config.getWidthChars() : textLabel().length() + 3;
   }
   
   /**
    * Set new value for the COLON option.
    * 
    * @param   colon
    *          New value for the option column.
    */
   @Override
   public void setColon(int colon)
   {
      config.setColon(colon);
   }
   
   /**
    * Assign new instance of <code>Label</code> to the widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label label)
   {
      this.label = label;
      this.label.setVisible(isVisible());
   }

   /**
    * Force NO-LABELS flag.
    * 
    * @param   labels
    *          New value of the flag.
    */
   @Override
   public void setNoLabels(boolean labels)
   {
      config.setNoLabels(labels);
   }

   /**
    * Set the new text for the component.
    * 
    * @param   text
    *          New text value.
    */
   public void setText(String text)
   {
      setTextLabel(text);

      repaint();
      
      if (config.isAutoResize())
         config.setWidthCharsDirect(textLabel().length() + 3);

   }

   /**
    * Set toggle-box label.
    * 
    * @param   label
    *          Label text.
    */
   public void setTextLabel(String label)
   {
      this.textLabel = label;
   }

   /**
    * Assign new TO column.
    * 
    * @param   col
    *          New value for the option.
    */
   @Override
   public void setTo(int col)
   {
      config.setTo(col);
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * This is called when the toggle-box takes the input focus. When the help string is defined
    * for the toggle-box it should be displayed at the window status line.
    *
    * @param   event
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      widgetFocused = true;
      
      // even for consecutive FOCUS_GAINED events, the HELP text must be set
      // for this toggle-box, because the HELP message-text may have been changed
      // for another widget (i.e. on a nested UPDATE)
      
      // draw the help string for focused toggle-box
      ThinClient.setHelp(config.getHelp());
   }
   
   /**
    * This is called when the button looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      widgetFocused = false;
      
      // restore the cursor visibility that existed when we gained focus 
      screen().setCursorStatus(curvis);
   }

   /**
    * This method is invoked when key is pressed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
   }

   /**
    * This method is invoked when key is typed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyTyped(KeyInput event)
   {
   }
   
   /**
    * Processes key events occurring on this object by dispatching them
    * to any registered KeyListener objects.
    *
    * @param    keyEvent
    *           The event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      // use arrow keys to move focus between widgets (only left/right work
      // in the alert box case)
      if (parent() instanceof AlertBox)
      {
         setActionCode(keyEvent);
      }
      else
      {
         if (handleCursorKeys(keyEvent.actionCode(), (Frame) UiUtils.locateFrame(this)))
            return;
      }

      handleBasicEvents(keyEvent);
   }

   /**
    * Handle cursor movement keys.
    * 
    * @param   key
    *          Key code.
    * @param   frame
    *          Parent frame.
    * 
    * @return  <code>true</code> if any cursor key was found.
    */
   private boolean handleCursorKeys(int key, Frame frame)
   {
      boolean returnNow = false;
      
      if (key == Keyboard.KA_TAB)
      {
         returnNow = nextFocus();
      }
      else if (key == Keyboard.KA_BACK_TAB)
      {
         returnNow = prevFocus();
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         // the left most widget ignores this condition
         if (!frame.isLeftMostWidget(this))
         {
            returnNow = prevFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_UP)
      {
         returnNow = moveAboveWidget(frame);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         // the right most widget ignores this condition
         if (!frame.isRightMostWidget(this))
         {
            returnNow = nextFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         returnNow = moveBelowWidget(frame);
      }
      return returnNow;
   }

   /**
    * Move focus to widget below in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveBelowWidget(Frame frame)
   {
      frame.moveBelowWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to widget above in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveAboveWidget(Frame frame)
   {
      frame.moveAboveWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to previous widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean prevFocus()
   {
      ThinClient.previousFocus(this);
      refresh();
      return true;
   }

   /**
    * Move focus to next widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean nextFocus()
   {
      ThinClient.nextFocus(this);
      refresh();
      return true;
   }

   /**
    * Refresh widget at the screen.
    */
   private void refresh()
   {
      draw();
      requestSync();
   }

   /**
    * Prepare event for processing.
    *   
    * @param   ki
    *          Input event.
    */
   private void setActionCode(KeyInput ki)
   {
      int key = ki.actionCode();
      // alert box processing has no need to generate entry/leave events 
      // so we take a very simple approach
      if (key == Keyboard.KA_CURSOR_LEFT)
      {
         ki.setActionCode(Keyboard.KA_BACK_TAB);
         ki.setKeyCode(Key.VK_BACK_TAB);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         ki.setActionCode(Keyboard.KA_TAB);
         ki.setKeyCode(Key.VK_TAB);         
      }
   }
   
   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   private void handleBasicEvents(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      if (key == '\t')
      {
         parent().nextFocus();
         return;
      }
      else if (key == Key.VK_BACK_TAB)
      {
         parent().prevFocus();
         return;
      }

      if (super.isEnabled() && (key == Key.VK_ENTER || key == ' '))
      {
//         EventManager.postEvent(new ItemEvent(this));
         // inverse the CHECKED state
         config.setChecked(!config.isChecked());
         refresh();
      }
   }
   
   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean wasVisible()
   {
      return true;
   }
   
   /**
    * Remember the widget was made visible at least once.
    */
   @Override
   public void setWasVisible()
   {
      return;
   }
}
